/*************************************************/
/* Author: Kyri Pavlou         5 October 2007    */
/*************************************************/

#include "algorithms.h"

/* global variables */
int *candidates;       /* this array plays a triple role: first it is used in the create_rightmost function; then it holds
		          the sequence of indices of the aux array that when traversed would yield the sorted candidate set 
			  elements. In the end it holds the sorted candidate set elements. */
int *aux;              /* this array holds temporarily the unsorted candidate set elements 
			  which are created by the generate function */
int tail_index = 0;    /* this maintains the logical size of the aux array */
int logical_size = 0;  /* logical size of candidates array */
int z=0;               /* this is the number of zeros in the target binary number */ 
int *Zarray;           /* this Zarray[i] gives you the the number of zeros seen to the left of target_number[i] inclusive */

CacheRec *cache;       /*the head of the linked list used to simulate the cache*/

/*****************************************************************/
/* Input: t is the target bit number used to generate		 */
/*	    the candidate set. 					 */
/*   length is the length of the bit representation of the       */
/*	    target number t					 */ 
/*	  k is dimension of the boolean domain of the bitwise    */
/*	    AND function  					 */
/* This function checks if the an already cached candidate set   */
/* can be used to generate the candidate set for target t.       */
/* This optimization is possible only if t is a suffix of a      */
/* target y whose candidate set is already computed and cached.  */
/* RT: O(cache size).			                         */
/*****************************************************************/
void candidateSetCached(unsigned int t, int length, int k)
{
    int t_start = -1;
    tail_index = z= logical_size = 0; /*initialize global variables*/
    CacheRec *ptr; /*linked list "iterator"*/
    ptr = cache;
    /*Check if the a cached result can be used instead
      of generating the candidate set from scratch */
    while(ptr != NULL)
    {
        DBG(printf("Traversing cache: ptr->l_y = %d , k = %d, num of zeros target %d=%d , ptr->Zarray[0] = %d , ptr->C[0] = %d\n", 
		   ptr->l_y, k, ptr->y, ptr->Zarray[length-1],  ptr->Zarray[0], ptr->C[0]));

        if( (ptr->y & t) == t ) /*checking if t is a suffix of y*/
        {
            t_start = ptr->l_y - length; /*note the start position of suffix t in string y*/
            break;
        }
        ptr = ptr->next;
    }
    /*if a cached result can be used to generate the candidate set the special function 
       candidateSetSuffix is called to do this*/
    if((t_start >= ZERO) && (k <= ptr->k)) 
    {
        candidateSetSuffix(t_start, ptr->l_y, k, ptr->Zarray[length-1], ptr->Zarray, ptr->C);
    }
    else /*otherwise the candidate set is created from scratch*/
	candidateSet(k, t, length);

    bzero(aux, tail_index);
    bzero(Zarray, length);
    free(Zarray);
    free(aux);
    //free(candidates);
    tail_index = z = 0;
}

/*****************************************************************/
/* Input: k is dimension of the boolean domain of the bitwise    */
/*	    AND function  					 */
/*  target_number is the target bit number used to generate      */
/*	    the candidate set. 					 */
/*   length is the length of the bit representation of the       */
/*	    target number t					 */ 
/* This function creates and outputs the candidate set of the    */
/* original target_number. 					 */
/*****************************************************************/
void candidateSet(int k, unsigned int target_number, int length)
{
    printf("inside the candidateset fucntion with k=%d, target=%d, length=%d\n", k, target_number, length);
    int i, j; 
    /*find z the number of zeros in the input target_number*/
    /*at the same time create the Zarray: Zarray[i] gives 
    the number of zeros seen to the left of target_number[i] inclusive*/
    Zarray = (int *)mymalloc(length * sizeof(int));
    for(i = length-1; i >= 0; i--)   /*O(length)*/
    {
        if((target_number & (1<<i)) == 0)
            z++;
        Zarray[length-i-1] = z;
    }	  

    /*Print the elements of the Z array for debug purposes*/
    DBG(printf("z is %d\n",z));
    DBG(printf("the elements of the Zarray = {")); 
    for(j = 0 ; j < length ; j++)
    {
        DBG(printf(" %d ",Zarray[j]));
    }
    DBG(printf("}\n")); 

    candidates = (int *)mymalloc((1<<z)*sizeof(int));
    aux = (int *)mymalloc((1<<z)*sizeof(int));

    if ((k < 1) || (k > (1 << length) ))
        candidates[0] = NOT_DEFINED;
    else if (k == 1)
    {
        logical_size = 1;
        candidates[0] = target_number;
    }
    else if ( (z == 0 && k > 1) || (length >= z && z > 0 && k > (1 << z) ) )
        logical_size = 0;
    else if (length >= z && z >0 && 2 <= k && k <= (1 << z))
    {
        logical_size = 1 << z;
        /*preprocessing required to answer "rightmost zero" query in constant time*/
        create_rightmost(target_number, length); 	     /* O(length) */      
        generate(target_number, candidates[length], length); /* generate the candidate set elements recursively in O(2^z) time*/
        funky_sort();                                        /* sort the generated candidate set elements O(2^z) */
      
        /*****************************************************************************************************/
        /*Update the cache with the newly created candidate set - add at the front -- no ordering exists     */
        /*CacheRec *newRec;
        newRec = (CacheRec *)mymalloc(sizeof(CacheRec));
        newRec->y = target_number;
        newRec->l_y = length;
        newRec->k = k;
        newRec->C = candidates;
        newRec->Zarray = Zarray;
        newRec->next = cache;
        cache = newRec;*/ 
	/*Not done here since it is useful across different databases and not within a database               */
	/*This is because two target bit patterns for one to be a proper suffix of the other implies          */
	/*that the number of chains in the tile differs between the two. In turn this implies that            */
        /*the initial seetings of the two tiled-bitmap algorithms which produced those target numbers differ. */
        /******************************************************************************************************/
    }
    if(candidates[0] == NOT_DEFINED)
        printf("Candidate set not defined\n");
    else
    {
        /*Print the candidate array contents*/			
        printf("The Candidate set C_{%d,%d} = { ", target_number, k);
        for(i=0; i < logical_size; i++)         /* O(2^z) */
            printf("%d ", candidates[i]);
        printf("}\n");
    }
}

/**************************************************************************/
/* Input: suffix_start is the index at which the new target      	  */ 
/*	  	whose candidate set is to be determined occurs as a    	  */ 
/*	 	suffix in the target whose candidate set is cached        */
/*   length is the length of the bit representation of the                */
/*	    cached target number					  */ 
/*	  k is dimension of the boolean domain of the bitwise    	  */
/*	    AND function  					          */
/*	  z is the number of zeros in the cached binary target            */
/*   Zarray is the Z array of the cached binary target. Zarray[i] gives   */
/*     	    the number of zeros seen to the left of                       */
/*	    target_number[i] inclusive                                    */
/* candidates is the cached candidate set				  */
/* Output: the candidate set of the suffix target number 		  */
/* This function creates the candidate set of the target number by using  */
/* the cached candidate set of another target. It does so by taking a     */
/* subset of appropriate cardinality from the cached candidate set and    */
/* applying a mask to each element.		  			  */
/**************************************************************************/
void candidateSetSuffix( int suffix_start, int length, int k, int z, int *Zarray, int *candidates)
{
    /*  Note that this entire if clause, the "for" loop in the last else clause notwithstanding, runs in constant time
    Also, the original target_number and its candidate set are not altered in any way*/
    if(suffix_start != 0)  /* if the suffix is the same as the original number there's no need to execute this */
    {
        int max;     /* the max element in the suffix candidate set */
        int i;          
        int suffix_length = length - suffix_start;   
        int z_suffix = z - Zarray[suffix_start-1] ;   /* z_suffix  is the number of zeros in the suffix */

	/* this mask of the from  000...01111..1 (with suffix_length number of '1's) 
	is used to take the "suffix" of a binary number*/
        unsigned int mask =  (1 << suffix_length)-1;   

        if ( (k < 1) || (k > (1 << suffix_length)) )
            candidates[0] = NOT_DEFINED;
        else if (k == 1)
            max = candidates[0] & mask;  /* no need to set logical_length since it was set above -- this is because k is not variable; 
					also, the max candidate element is the only element */
        else if ( (z_suffix == 0 && k > 1 && k <= (1 << suffix_length)) || (suffix_length >= z_suffix && z_suffix > 0 && k > (1 << z_suffix) ) )
            logical_size = 0;                      /* the candidate set is empty */
        else if (suffix_length >= z_suffix && z_suffix > 0 && 2 <= k && k <= (1 << z_suffix))
            max = candidates[(1 << z_suffix) - 1] & mask;     /* find the max element and take its suffix */
	
        if(candidates[0] == NOT_DEFINED)
            printf("Candidate set not defined\n");
        else if (logical_size == 0)
            printf("The Candidate set is empty\n");
        else
        {
            printf("The Candidate set given explicitly :\nC_{%d,%d} = { ", candidates[0] & mask, k); /*output the suffix candidates set explicitly*/
            for(i = 0; i < (1<<z_suffix); i++)            /* O(2^z_suffix) */
                printf("%d ", candidates[i] & mask);
            printf("}\n");

            printf("The mask and the  q-1 (the index of the last element in the candidate set for suffix) are:\n mask=%d q-1=%d \n", mask , (1 << z_suffix)-1);  
            /*can also define the suffix candidate set by giving the two defining numbers q and mask*/
        }
    }
}


/*******************************************************************/
/* Input: i: is the modified bit pattern at each stage of the 	   */
/*	     recursion 						   */ 
/*	  p: the position of zeros in the target binary number;    */
/*	     used to guide the recursion from right to left on the */
/*	     binary target.	    				   */
/*   length: is the length of the bit representation of the        */
/*	     target number.					   */
/* Output: the aux array contains the candidate granules           */
/* This function recursively computes the candidate set elements   */
/* (albeit unsorted). At each position in the bit pattern where    */
/* there is a 0 it sets it to 1 and recurses in order to enumerate */ 
/* all strings with (0/1) at each position where in the target     */ 
/* a 0 was found. RT: O(2^z) 				           */
/*******************************************************************/
void  generate(unsigned int i, int p, int length)
{
    if (p == GENERATE)
    {
        aux[tail_index] = i; /*keep the generated elements in the aux array*/
        tail_index++;
    }	
    else
    {
        /* the candidate set at this point holds the indices of the rightmost 0 to the left of given position*/
        int pos = candidates[p];  /*find index of rightmost 0 to the left of position p */
        generate(i,pos, length);
        generate(i+(1<<(length - p - 1)),pos, length);  /*set the 0 to a 1 and recurse*/
    }
}

/*****************************************************************/
/* Input: target_number is the target bit number used to generate*/
/*	   the candidate set. 					 */
/*	  length is the length of the bit representation of the  */
/*	  target number						 */
/* Output: the populated candidates array			 */
/* This function performs pre-processing using the target_number */
/* and fills the candidates array in order to answer the         */
/* "rightmost zero" query in constant time. More specifically,   */ 
/* candidates[p] is the index of the rightmost zero to the left  */
/* of index p NON-INCLUSIVE. RT: O(length).                      */
/*****************************************************************/
void create_rightmost(unsigned int target_number, int length)
{
    int i, j, flag, jj;
    j = GENERATE;
    /* the flag is required because we must remember 
    what we saw in the previous iteration; 
    if flag = TRUE we saw a 0 otherwise we saw a 1*/
    flag = FALSE;   
    for(i = length-1; i >= -1; i--)  
    {	
        if(flag) 
            j = length-i-2;
        if((target_number & (1 << i)) == 0)
            flag = TRUE;
        else 
            flag = FALSE;
        candidates[length-i-1] = j;
    }
    /*Print contents of candidates array for debug purposes*/
    DBG(printf("the candidates array ={"));
    for(jj  = 0 ; jj < length ; jj++)
    {
        DBG(printf(" %d ", candidates[jj]));
    }	
    DBG(printf("}\n"));
}

/***************************************************************/
/* This function sorts the candidate set elements. RT: O(2^z)  */
/***************************************************************/
void funky_sort(void)
{
    candidates[0] = 0; /*the 0th index is always the first one to be visited*/
    int i, offset, power;
    offset = 0;
    power = 1 << z;
    /*this algorithm performs a single pass over the candidates array and creates each new index by manipulating appropriately  previous ones*/
    for(i = 1; i < (1 << z); i++)                            /*O(2^z) time*/
    {
        /*if i is a power of 2 reset offset and adjust power*/
        if((i & (i - 1))== 0)   /* even though this test accepts 0 as a power of 2 in this case this is exactly what we want!!!*/
        {
            power = power >> 1; 
            offset = 0;
        }
	/*create the sequence of indices which when used to
	 index in the aux array will sort the candidate set elements*/
        candidates[i] = candidates[offset] +  power;
        offset++;
    }
    for(i = 0; i < (1 << z); i++)            /*O(2^z) time*/
        candidates[i] = aux[candidates[i]];  /*this is where the sorting happens*/   
}







