/*************************************************/
/* Author: Kyri Pavlou         24 September 2007 */
/*************************************************/

#include "algorithms.h"

/*3-D array of partial hash chains of the binary tree used by a3D algorithm*/
extern int ***P;
/*counter for number of times the notarization service
is contacted during the forensic analysis phase of the a3D algorithm*/
extern int counter_a3D_f; 

int *C_set_a3D;   		/*the candidate set*/
int c_a3D_capacity = INIT_CAP; 	/*the capacity of the candidate set*/
int c_a3D_logical = ZERO; 	/*the logical size of the candidate set*/

/*************************************************************/
/* The a3D function calls the recursive function aD3_helper()*/
/* in order to generate the candidate set.	   	     */
/*************************************************************/
void a3D()
{
    int j, min, LTB, UTB;
    /*the (maximum) height of the complete tree*/
    int m_max = 1 << (int)ceil(log_base_n(t_FVF/I_V, 2));
    /*height is the maximum number-1 of levels the complete binary tree; level_max-1*/
    int height =  (int)ceil(log_base_n(N,2)) + (int)floor(log_base_n(m_max,2)); 

    C_set_a3D = (int *)mymalloc(c_a3D_capacity * sizeof(int));

    DBG(printf("t_FVF = %d, I_V = %d , N = %d\n", t_FVF, I_V, N));
    DBG(printf("Before calling a3D helper\n"));
    DBG(printf("Entering recursion at P[%d, %d, %d] \n", m_max, height, 0));

    a3D_helper(m_max-1, height, 0);  /*call to the recursive function*/

    /*set the temporal bounds on CE*/
    min = C_set_a3D[0];
    if(min < t_FVF - I_V)
        LTB = t_FVF - I_V;
    else
        LTB = min;
    UTB = t_FVF;
    printf("%d < t_c <= %d\n", LTB, UTB);
    printf("****The C_set is****\n");
    printf("C_set = {");
    for(j = 0; j < c_a3D_logical; j++)
    {
        if(R_s != 1) /*if R_s is not 1 then give ranges of possible corrupted time granules*/
            printf(" [%d,%d] ", C_set_a3D[j]+1, C_set_a3D[j] + R_s - 1 + 1);
        else
            printf(" %d ", C_set_a3D[j] + 1);
    }
    printf("}\n\n");
    free(C_set_a3D);
}

/***************************************************************/
/* The aD3_helper function recursively visits every hash chain */
/* in the binary tree in order to generate the candidate set.  */
/* At each node/hash chain the tree it checks if the chain     */ 
/* exists in the complete tree or not. If it does not then     */ 
/* it recurses on to the children. If it exists, then it checks*/  
/* the validation result for the chain. If it is FALSE then it */
/* recurse onto the children; if it is TRUE it just returns    */
/*Input:  m: the height of the (sub)tree rooted at the current */
/*	     node/chain					       */
/*    level: the current level in the group of chains at       */
/*	     created at each VE 	   		       */
/*component: the position of the chian within a particular     */
/*           level  					       */
/*Ouput: C_set_a3D, a set containing all potentially corrupted */
/*       granules					       */
/***************************************************************/
void a3D_helper(int m, int level, int component)
{
    DBG(printf("Recursive step: on chain P[%d, %d , %d]\n", m+1, level, component));
    if((level == 0)) /*reached the leaves of the tree*/
    {
        if (exists(m))
        {
            if(val_check_a3D(m, level, component) == FALSE)
            {
                append_a3D(&C_set_a3D, component*R_s);
                c_a3D_logical++;
            }
        }
        return;
    }
    else
    {
        if( exists(m) == FALSE || (val_check_a3D(m, level, component) == FALSE))
        {
            if(!(level <= log_base_n(N,2)) ) //recurse on completing chains(i.e. chains which are in addition to ones that form the tile)
            {
                a3D_helper((m + m-((1 << level)/N))/2, level-1, 2*component); //recurse to the left child
                a3D_helper(m, level-1, 2*component+1); //recurse to the right child
            }
            else if(level <= log_base_n(N,2)) //recurse within tile chains
            {
                a3D_helper(m, level-1, 2*component); //recurse to the left child
                a3D_helper(m, level-1, 2*component+1); //recurse to the right child
            }
        }
    }
}

/***************************************************************/
/* Appends corrupted granule g to candidate set C_set_a3D      */
/* If there is not enough space in the set the array is        */
/* resized to double its previous size.                        */
/***************************************************************/
void append_a3D(int **C_set_a3D, int g)
{
    DBG(printf("Appending %d to candidate set...\n", g));
    if(c_a3D_logical < c_a3D_capacity)
    {
        (*C_set_a3D)[c_a3D_logical] = g;
    }
    else
    {
        DBG(printf("Resizing the candidate set...\n"));
        int i;
        int *C_temp;
        c_a3D_capacity *= 2;
        C_temp = (int *)mymalloc(c_a3D_capacity*sizeof(int));
        for(i = 0; i < c_a3D_logical; i++)
        {
            C_temp[i] = (*C_set_a3D)[i];
        }
        C_temp[c_a3D_logical] = g;
        free(*C_set_a3D);
        *C_set_a3D = C_temp;
    }
}

/******************************************************************/
/* Returns the validation result (T/F) of the tree node           */  
/* hash chain P[m][level][component]. For certain hash chains     */
/* (height of substree is a power of 2, and chain is the first of */
/* those not belonging to a tile) the validation result is        */
/* retrieved from the BlackChains array.                          */
/******************************************************************/
int val_check_a3D(int m, int level , int component)
{
    counter_a3D_f++;

    if( ((m+1) & m) == 0  &&  level >= (int)log_base_n(N,2) && component == 0)
    	return BlackChains[m];
    else
    	return P[m][level][component];
}

/********************************************************/
/* Checks if chain with subtree rooted there and  	*/
/* height m exists in the complete tree.		*/
/* Returns TRUE if the chain exists, FALSE otherwise.	*/
/********************************************************/
int exists(int m)
{
    if(m >= t_FVF/I_V)
        return FALSE;
    else
        return TRUE;
}



