/**********************************************/ 
/*                     			      */ 
/* Author: Kyri Pavlou  		      */
/* 4 October 2007             		      */ 
/*                     			      */ 
/**********************************************/

#ifndef __ALG__
#define __ALG__

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <strings.h>

#define FALSE        (0)
#define TRUE         (1)
#define NOT_DEFINED (-9)
#define NOT_ADDED   (-6)
#define GENERATE    (-1)
#define ZERO         (0)

#define INIT_CAP     (4) /* Macros for command line arguments*/
#define MIN_NUM_ARGS (7)
#define FLAG_SIZE    (1)
#define RT_INDEX     (1)
#define ALG_INDEX    (2)
#define N_INDEX      (3)
#define V_INDEX      (4)
#define TFVF_INDEX   (5)
#define TL_START     (6)

#define RED	     (2)
#define GREEN        (1)
#define BLUE         (0)
#define YELLOW       (3)
#define COLOR_NUM    (4)

/*Macro to calculate the maximum of two numbers x and y*/
#define MAX(x,y) ((x) < (y) ? (y) : (x))

/*Debug macro for verbose output*/
#ifdef DEBUG
#define DBG(X) \
{ printf("[ %s:%d ]  ",__FILE__, __LINE__); ((X)); }
#else
#define DBG(X) \
{}
#endif

/*Struct of records which make up the cache*/    
typedef struct cache_rec{
unsigned int y ; 	    	   /*the cached target*/
int l_y; 		     	   /*the length of the target*/
int k; 				   /*the parameter for which the candidate set was computed; the dimension of the boolean domain*/
int *C; 			   /*the candidate set*/
int *Zarray; 			   /*the Z array of zeros for target y*/
struct cache_rec *next; 	   /*pointer to the next element in the linked list*/
}CacheRec;

/*Chain Struct used to represent partial chains in the RGBY algorithm*/    
typedef struct chain{
int color;		
int subscript;		/*this is referred to as the "number" of the partial chain in the TODS paper*/
}Chain;

extern int *BlackChains;           /*integer array simulating the black cumulative chains*/
extern int *RGBYchains;	           /*integer array simulating the RGBY partial chains*/
extern int *tiles;                 /*integer array simulating the partial chains in the tiles*/
extern CacheRec *cache;		   /*the head for the linked list cache*/
extern int cache_length;	   /*the logical size of the cache*/

/*the notarization interval*/ 
/*the validation factor*/
/*the notarization factor*/
/*the validation interval*/ 
/*the time of first validation failure*/
/*the spatial detection resolution*/
/*the number of partial chains within a tile*/
extern int I_N, V, N, I_V, t_FVF, R_s, num_chains_per_tile;

/*function signatures for CandidateSet_Fastest.c*/
void  generate(unsigned int i, int pos, int length);
int find_rightmost(unsigned int i, int p);
void * mymalloc(int size);
void funky_sort(void); 
void create_rightmost(unsigned int target_number, int length);
void candidateSet(int k, unsigned int target_number, int length);
void candidateSetSuffix( int suffix_start, int length, int k, int z, int *Zarray, int *candidates);
void candidateSetCached(unsigned int t, int l, int k);

/*function signatures for Tiled_Bitmap.c*/
void Tiled_Bitmap(int t_FVF, int I_N, int k); 
int val_check(int t, int i);
void append(int **C_set, int g);

/*function signatures for forensic.c*/
float log_base_n(float num, float base);
int find_t_RVS(int t_FVF, int I_N);

/*function signatures for  monochromatic.c*/
void monochromatic(int t_FVF, int I_N, int V);

/*function signatures for RGBY.c*/
void RGBY(int t_FVF, int I_N, int R_s);
void append_RGBY(int **C_set_RGBY, int g);
void successor(Chain **succChain, Chain *currChain);
int val_check_RGBY(int t_FVF, int I_V, Chain *test);
int less_or_equal(Chain *chain1, Chain *chain2);
char number_to_color(int num);

/*function signatures for a3D.c*/
void a3D();
void a3D_helper(int m, int level, int component);
int val_check_a3D(int m, int level , int component);
int exists(int m);
void append_a3D(int **, int g);


#endif
