/*************************************************/
/* Author: Kyri Pavlou        			 */
/* 24 September 2007 				 */
/*************************************************/

#include "algorithms.h"
/***************************************************/
/* Input                                           */
/*  t_FVF: the time of first validation failure    */
/*  I_N: notarization interval                     */
/*  V: validation factor                           */
/* Output                                          */
/*  bounds on t_c and t_l                          */
/*  USB upper spatial bound                        */
/*  LSB lower spatial bound                        */
/*  UTB upper temporal bound                       */
/*  LTB lower temporal bound                       */
/***************************************************/

void monochromatic( int t_FVF, int I_N, int V)
{
    int  I_V, t_RVS, USB, LSB, UTB, LTB;
    I_V = V * I_N; 
    t_RVS = find_t_RVS(t_FVF, I_N);  /*find time of most recent validation success*/
    USB = t_RVS + I_N;
    LSB = t_RVS;
    UTB = t_FVF;
    if((t_FVF - I_V) < t_RVS)   /*if the CE is introactive, i.e., within the latest I_V*/
        LTB = t_RVS;		/*it is (temporally) bounded below by t_RVS*/
    else
        LTB = t_FVF - I_V;
    DBG(printf("t_RVS = %d, LSB = %d, USB = %d, LTB = %d, UTB = %d\n", t_RVS, LSB, USB, LTB, UTB));
    if(t_RVS >= (t_FVF - I_V))
    {
        printf("Introactive CE\n");
    }
    else if(t_RVS < (t_FVF - I_V))
    {
        printf("Retroactive CE\n");
    }
    printf("Data Only CE, %d < t_l <= %d, %d < t_c <= %d\n", LSB, USB, LTB, UTB);
    printf("Postdating CE, %d < t_l <= %d, %d < t_c <= %d, %d < t_p <= %d\n", LSB, USB, LTB, UTB, USB, t_FVF);
    printf("Backdating CE, %d < t_b <= %d, %d < t_c <= %d, %d < t_l <= %d\n", LSB, USB, LTB, UTB, USB, t_FVF);
}






